import requests  # assuming you're using the requests library to make HTTP requests
from typing import Dict, Any
import promptlabs.chain as Chain
from promptlabs.pipeline import Pipeline


class Client:
    def __init__(self, api_key):
        self.api_key = api_key
        self.endpoint_url = "https://pops.up.railway.app/api/package"


    def _fetch(self, url):
        try:
            response = requests.get(
                self.endpoint_url + url,
                headers={"Authorization": "Bearer " + self.api_key},
            )
        except Exception as exc:
            raise Exception("Could not fetch data from API") from exc

        if response.status_code != 200:
            raise Exception("Could not fetch data from API")
    
        return response
    
    def chain(self, chain_name) -> Chain:
        data = self._fetch(
            f"/chain/{chain_name}",
        ).json()
        
        print(data)
        
        try:
            chain_class = getattr(Chain, data['type'])
            print(chain_class)
            return chain_class(config=data)
        except Exception as e:
            print(e)
            raise ValueError(f"Invalid chain type: {data['type']}")
        
    def chain_init_data(self, chain_name) -> Dict[str, Any]:
        data = self._fetch(
            f"/chain/{chain_name}",
        ).json()
        
        return data
        
    def pipeline(self, pipeline_name) -> Pipeline:
        data = self._fetch(
            f"/pipeline/{pipeline_name}",
        ).json()
        
        # get the chain configs
        chains_data = self._fetch(
            f"/pipeline_chains/{pipeline_name}"
        ).json()
        
        data['chain_configs'] = chains_data
        
        return Pipeline(config=data)
    
    def pipeline_init_data(self, pipeline_name) -> Dict[str, Any]:
        data = self._fetch(
            f"/pipeline/{pipeline_name}",
        ).json()
        
        # get the chain configs
        chains_data = self._fetch(
            f"/pipeline_chains/{pipeline_name}"
        ).json()
        
        data['chain_configs'] = chains_data
        return data

def create_client(api_key):
    return Client(api_key)



import warnings
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union
from abc import ABC, abstractmethod
from datetime import datetime

from pydantic import Extra, Field, root_validator, validator

import promptlabs.chain as chain
from promptlabs.chain.base_chain import Chain
from promptlabs.scheme.template import ChainOutput, ChainLog, PipelineLog, PipelineOutput

def create_chain(chain_config) -> Chain:
    try:
        chain_class = getattr(chain, chain_config['type'])
        return chain_class(config=chain_config)
    except:
        raise ValueError(f"Invalid chain type: {chain_config['type']}")
class Pipeline:
    
    def __init__(self, config : Dict[str, Any]):
        self.config = config
        # add id to each chain_config from chain_id
        for chain_config in self.config["chain_configs"]:
            print(chain_config)
            chain_config['id'] = chain_config['chain_id']
        
        self.chains = [create_chain(chain_config) for chain_config in self.config["chain_configs"]]
    
    @property
    def input_keys(self) -> List[str]:
        return self.config["input_keys"]
    
    @property
    def output_keys(self) -> List[str]:
        
        return self.config["output_keys"]
                
                
    @root_validator(pre=True)
    def validate_chains(cls, values: Dict) -> Dict:
        config = values["config"]
        chain_configs = config["chain_configs"]
        pipe_input_keys = config["input_keys"]
        
        known_variables = pipe_input_keys
        for i, chain_config in enumerate(chain_configs):
            missing_vars = set(chain_config["input_keys"]).difference(known_variables)
            if missing_vars:
                raise ValueError(
                    f"Missing required input keys: {missing_vars} at chain {i}, "
                    f"chain config with error : {chain_config},"
                    f"only had {known_variables}"
                )
            known_variables |= set(chain_config["output_keys"])
        
        missing_vars = set(config["output_keys"]).difference(known_variables)
        if missing_vars:
            raise ValueError(
                    f"Expected output variables that were not found: {missing_vars}."
                )
        
        return values
        
    def _validate_inputs(self, inputs: Dict[str, Any]) -> None:
        """Check that all inputs are present."""
        missing_keys = set(self.input_keys).difference(inputs)
        if missing_keys:
            raise ValueError(f"Missing some input keys: {missing_keys}")
        
    def _validate_outputs(self, outputs: Dict[str, Any]) -> None:
        """Check that all outputs are present."""
        missing_keys = set(self.output_keys).difference(outputs)
        if missing_keys:
            raise ValueError(f"Missing some output keys: {missing_keys}")

    
    def run(self, inputs: Dict[str, Any]):
        
        # input validate
        try:
            self._validate_inputs(inputs)
        except Exception as e:
            return (
                PipelineOutput(pipeline_success=False),
                PipelineLog(logs={"error_log" : str(e)})
            )
            
        # pipeline log
        pipeline_log = PipelineLog()
        
        # run chains
        known_variables = inputs
        
        is_fail = False
        for i, chain_in_pipe in enumerate(self.chains):
            chain_inputs = {key: val for key, val in known_variables.items() if key in chain_in_pipe.input_keys}
            chain_outputs = {key: val for key, val in known_variables.items() if key in chain_in_pipe.output_keys}
            if chain_in_pipe is not None and not is_fail:
                if chain_in_pipe._chain_type == "FunctionChain" and chain_in_pipe._function is not None:
                    chain_output, chain_log = chain_in_pipe.run(chain_inputs)
                else:
                    chain_output, chain_log = chain_in_pipe.run(chain_inputs,chain_outputs)
                
                pipeline_log.chain_logs.append(chain_log)
                pipeline_log.chain_outputs.append(chain_output)
                
                if chain_output.chain_success == False:
                    is_fail = True
                    error_message = "{chain_type} Chain Step#{i} failed. Error log: {error_log}".format(
                        chain_type=type(chain_in_pipe),
                        i=i+1, 
                        error_log=chain_log.error_log)
                    pipeline_log.logs['error_log'] = error_message
                else:
                    outputs = chain_output.output_variables
                    known_variables.update(outputs)
            else:
                pipeline_log.chain_logs.append(
                    ChainLog(logs={"error_log" : "Error occurs before start this chain"})
                )
                pipeline_log.chain_outputs.append(
                    ChainOutput(chain_success=False)
                )
        
        if is_fail:

            return (
                PipelineOutput(pipeline_success=False),
                pipeline_log
            )
        
        output_variables = {key: val for key, val in known_variables.items() if key in self.output_keys}       

        return (
            PipelineOutput(pipeline_success=True, output_variables=output_variables),
            pipeline_log
        ) 



import warnings
import os
from datetime import datetime
import re

from typing import Any, Dict, List, Optional, Sequence, Tuple, Union, Callable
from abc import ABC, abstractmethod

from pydantic import Extra, Field, root_validator, PrivateAttr 
from promptlabs.chain.base_chain import Chain
from promptlabs.scheme.template import ChainOutput, ChainLog

class FunctionChain(Chain):
    
    _function: Callable = PrivateAttr(default=None)
    
    def insert_function(self, function: Callable) -> 'FunctionChain':
        self._function = function
        return self

    def remove_function(self) -> None:
        self._function = None
        return
    
    @root_validator(pre=True)
    def set_function(cls, values: Dict) -> Dict:
        if 'function' in values['config']:
            function = values['config']['function']
        else:
            function = None
        values['_function'] = function
        return values
    
    @property
    def _chain_type(self) -> str:
        return self.config["type"]
    
    @property
    def _have_output(self) -> bool:
        if self.config['output_keys'] == []:
            return False
        else:
            return True
    
    def validate(
        self,
        inputs: Dict[str, Any],
        outputs: Dict[str, Any] = {}
    ) -> Dict[str, Any]:
        
        # input validate
        try:
            self._validate_inputs(inputs)
        except Exception as e:
            return {   
                "validated" : False,
                "error_log": str(e),
                "output_variables" : {}
            }
        
        # function validate
        if self._have_output:
            if self._function is None and outputs == {}:
                return {
                    "validated" : False,
                    "error_log": "You should Define Function or Mock outputs",
                    "output_variables" : {}
                }  
                
        if self._function is not None and outputs != {}:
            return {
                "validated" : False,
                "error_log": "You should Define Function or Mock outputs, not both",
                "output_variables" : {}
            }    
            
        if self._function:
            try:
                outputs = self._function(inputs)
            except Exception as e:
                return {
                    "validated" : False,
                    "error_log": str(e),
                    "output_variables" : {}
                }
            
        # output validate
        try:
            self._validate_outputs(outputs)
        except Exception as e:
            return {
                "validated" : False,
                "error_log": str(e),
                "output_variables" : {}
            }
        
        return {
            "validated" : True,
            "output_variables" : outputs
        }
    
    def run(
        self,
        inputs: Dict[str, Any],
        outputs: Dict[str, Any] = {}
    ) -> Tuple[ChainOutput, ChainLog]:
        
        # input validate
        try:
            self._validate_inputs(inputs)
        except Exception as e:
            return (
                ChainOutput(chain_success=False),
                ChainLog(logs = {"error_log": str(e)})
            )
            
        # function validate
        if self._function is None and outputs == {} and self.output_keys != []:
            return (
                ChainOutput(chain_success=False),
                ChainLog(logs = {"error_log": "You should Define Function or Mock outputs"})
            ) 
        
        if self._function is not None and outputs != {}:
            return (
                ChainOutput(chain_success=False),
                ChainLog(logs = {"error_log": "You should Define Function or Mock outputs, not both"})
            ) 
            
        if self._function is not None:
            try:
                outputs = self._function(inputs)
            except Exception as e:
                return (
                    ChainOutput(chain_success=False),
                    ChainLog(logs = {"error_log": str(e)})
                )
            
        # output validate
        try:
            self._validate_outputs(outputs)
        except Exception as e:
            return (
                ChainOutput(chain_success=False),
                ChainLog(logs = {"error_log": str(e)})
            )
            
        return (
                ChainOutput(chain_success=True, output_variables=outputs),
                ChainLog()
            )



import warnings
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union
from abc import ABC, abstractmethod

from pydantic import BaseModel, Extra, Field, root_validator, validator
from promptlabs.scheme.template import ChainOutput, ChainLog

class Chain(BaseModel, ABC):
    """Base class for all chains."""
    config : Dict[str, Any]
    
    @property
    def input_keys(self) -> List[str]:
        return self.config["input_keys"]
    
    @property
    def output_keys(self) -> List[str]:
        return self.config["output_keys"]
    
    @property
    def _chain_type(self) -> str:
        raise NotImplementedError("Saving not supported for this chain type.")
    
    def _validate_inputs(self, inputs: Dict[str, Any]) -> None:
        """Check that all inputs are present."""
        missing_keys = set(self.input_keys).difference(inputs)
        if missing_keys:
            raise ValueError(f"Missing some input keys: {missing_keys}")

    def _validate_outputs(self, outputs: Dict[str, Any]) -> None:
        missing_keys = set(self.output_keys).difference(outputs)
        if missing_keys:
            raise ValueError(f"Missing some output keys: {missing_keys}")
        
    # @abstractmethod
    # def validate(
    #     self,
    #     inputs: Dict[str, Any],
    #     outputs: Dict[str, Any] = {}
    # ) -> Dict[str, Any]:
    #     """Validate the chain."""
    #     pass
    
    # @abstractmethod
    # def test(
    #     self,
    #     inputs: Dict[str, Any],
    #     outputs: Dict[str, Any] = {}
    # ) -> Dict[str, Any]:
    #     """Run the chain."""
    #     pass
    
    @abstractmethod
    def run(
        self,
        inputs: Dict[str, Any],
        outputs: Dict[str, Any] = {}
    ) -> Tuple[ChainOutput, ChainLog]:
        """Run the chain."""
        pass


"""
promptlabs: A Python package for PromptOps (versioning, logging, experimenting, etc.)
"""
from setuptools import setup, find_packages

setup(
    name="promptlabs",
    version="0.0.2",
    packages=find_packages(),
    description="promptlabs: A Python package for PromptOps (versioning, logging, experimenting, etc.)",
    auther="weavel",
    install_requires=['openai', 'pydantic', 'requests'],
    python_requires='>=3.7.1',
    keywords=['weavel', 'prompt', 'llm', 'promptops', 'promptlabs', 'prompt engineering']
)



import warnings
import os
from datetime import datetime
import re

from typing import Any, Dict, List, Callable, Optional, Sequence, Tuple, Union
from abc import ABC, abstractmethod

from pydantic import Extra, Field, root_validator, validator

from promptlabs.chain.base_chain import Chain
from promptlabs.scheme.template import ChainOutput, ChainLog
from promptlabs.scheme.llm import ChatOpenAI, BaseLLM

class LLMChain(Chain):

    llm : BaseLLM
    
    @root_validator(pre=True)
    def set_llm(cls, values: Dict) -> Dict:
        if "llm_function" not in values['config']:
            chat_openai_config = {
                "model_name": values['config']['model_name'],
                "temperature": values['config'].get('temperature', 0.7),
                "max_tokens": values['config'].get('max_tokens')
            }
            if 'openai_api_key' in values['config']:
                chat_openai_config["openai_api_key"] = values['config']['openai_api_key']
            values['llm'] = ChatOpenAI(
                **chat_openai_config
            )
        else:
            values['llm'] = values['config']["llm_function"]
        return values
    
    @property
    def _chain_type(self) -> str:
        return self.config["type"]
    
    @property
    def model_name(self) -> str:
        return self.config["model_name"]
    
    @property
    def prompt(self) -> str:
        return self.config["prompt"]
    
    @property
    def system_prompt(self) -> str:
        default_system_prompt = "You are a helpful assistant."
        if 'system_prompt' not in self.config:  
            return default_system_prompt
        else:
            return self.config["system_prompt"]
    
    
    # def set_llm(self) -> BaseLLM:
    #     if "llm_function" not in self.config:
    #         if 'temperature' not in self.config:
    #             temperature = 0.7
    #         if 'max_tokens' not in self.config:
    #             max_tokens = None
    #         return ChatOpenAI(
    #             model_name=self.model_name,
    #             openai_api_key=self.config['openai_api_key'],
    #             temperature=temperature,
    #             max_tokens=max_tokens
    #         )
    #     else:
    #         return self.config["llm_function"]

    def _call_llm_submodule(
        self,
        inputs: Dict[str, Any]
    ) -> Dict[str, Any]:
        max_retry = 3
        retry_count = 0
        raw_output = None
        
        while retry_count < max_retry:
            openai_call_error = None
            parsing_error = None
            validation_error = None
            try:
                try:
                    messages = [
                        {"role": "system", "content" : self.system_prompt},
                        {"role": "user", "content" : self.prompt.format(**inputs)}
                    ]
                    is_success, response = self.llm.generate(
                        messages=messages
                    )
                    print(is_success, response)
                    if not is_success:
                        raise Exception("OpenAI call error : " + str(response))
                    raw_output = response["response"]['choices'][0]["message"]['content']
                except Exception as e:
                    openai_call_error = "OpenAI call error: " + str(e)
                    raise

                parsed_output = {
                    "text" : raw_output
                }

                try:
                    self._validate_outputs(parsed_output)
                except Exception as e:
                    validation_error = "Validation error: " + str(e)
                    raise

                return {
                    "status" : "success",
                    "parsed_output" : parsed_output,
                    "raw_output" : response,
                }
                
            except Exception as e:
                retry_count += 1
                print("retry count: " + str(retry_count))
                if retry_count == max_retry:
                    error_log = "\n".join(filter(None, [openai_call_error, parsing_error, validation_error]))

                    if raw_output:
                        return {
                            "status": "error", 
                            "error_log": error_log,
                            "raw_output": raw_output
                        }
                    else:
                        return {
                            "status": "error", 
                            "error_log": error_log
                        }
                continue
        
    
    def validate(
        self,
        inputs: Dict[str, Any],
        outputs: Dict[str, Any] = {}
    ) -> Dict[str, Any]:
        """Validate the chain."""
        
        # input validate
        try:
            self._validate_inputs(inputs)
        except Exception as e:
            return {   
                "validated" : False,
                "error_log": str(e),
                "output_variables" : {}
            }
            
        # openai call validate    
        llm_result = self._call_llm_submodule(inputs)
        
        if llm_result['status'] == 'error':
            return {   
                "validated" : False,
                "error_log": llm_result['error_log']
            }
        else:
            return {
                "validated" : True,
                "output_variables" : llm_result['parsed_output']
            }

    
    def run(
        self,
        inputs: Dict[str, Any],
        outputs: Dict[str, Any] = {}
    ) -> Tuple[ChainOutput, ChainLog]:
        
        # Validate inputs
        try:
            self._validate_inputs(inputs)
        except Exception as e:
            return ChainOutput(chain_success=False), ChainLog(logs={"error_log": str(e)})

        # openai call validate
        llm_result = self._call_llm_submodule(inputs=inputs)

        if llm_result['status'] == 'error':
            return (
                ChainOutput(chain_success=False),
                ChainLog(logs={"error_log": llm_result['error_log']})
            )
        else:
            return (
                ChainOutput(chain_success=True, output_variables=llm_result['parsed_output']),
                ChainLog(logs={"raw_output": llm_result['raw_output']})
            )

class LLMRegexChain(LLMChain):
    def _call_llm_submodule(
        self,
        inputs: Dict[str, Any]
    ) -> Dict[str, Any]:
        max_retry = 3
        retry_count = 0
        raw_output = None
        
        while retry_count < max_retry:
            openai_call_error = None
            parsing_error = None
            validation_error = None
            try:
                try:
                    messages = [
                        {"role": "system", "content" : self.system_prompt},
                        {"role": "user", "content" : self.prompt.format(**inputs)}
                    ]
                    is_success, response = self.llm.generate(
                        messages=messages
                    )
                    print(is_success, response)
                    if not is_success:
                        raise Exception("OpenAI call error : " + str(response))
                    raw_output = response["response"]['choices'][0]["message"]['content']
                except Exception as e:
                    openai_call_error = "OpenAI call error: " + str(e)
                    raise

                parsed_output = {}

                try:
                    output_keys = self.output_keys
                    for key in output_keys:
                        # capitalize the first letter of the key
                        capitalized_key = key[0].upper() + key[1:]
                        # make the regular expression
                        pattern = r"\[{}(\s*\(.+\))?\sstart\](.*?)\[{}(\s*\(.+\))?\send\]".format(capitalized_key, capitalized_key)
                        results = re.findall(pattern, raw_output, re.DOTALL)
                        results = [result[1] for result in results]
                        if results:
                            if len(results) > 1: 
                                raise Exception("Multiple Matches")
                            parsed_output[key] = results[0].strip()
                except Exception as e:
                    parsing_error = "Parsing error: " + str(e)
                    raise

                try:
                    self._validate_outputs(parsed_output)
                except Exception as e:
                    validation_error = "Validation error: " + str(e)
                    raise

                return {
                    "status" : "success",
                    "parsed_output" : parsed_output,
                    "raw_output" : response,
                }
                
            except Exception as e:
                retry_count += 1
                print("retry count: " + str(retry_count))
                if retry_count == max_retry:
                    error_log = "\n".join(filter(None, [openai_call_error, parsing_error, validation_error]))

                    if raw_output:
                        return {
                            "status": "error", 
                            "error_log": error_log,
                            "raw_output": raw_output
                        }
                    else:
                        return {
                            "status": "error", 
                            "error_log": error_log
                        }
                continue
        